####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np
from _SpecBase import inputFileDelim

####################################################################################################################################
#### SpecDRAM specification type-checking class
####################################################################################################################################

def adaptiveUpdateCount(self,adaptiveUpdateCount):
    if isinstance(adaptiveUpdateCount,int):
        return "adaptiveUpdateCount=" + str(adaptiveUpdateCount) + inputFileDelim
    else:
        raise TypeError("The input specification, adaptiveUpdateCount, must be of type int.")

####################################################################################################################################

def adaptiveUpdatePeriod(self,adaptiveUpdatePeriod):
    if isinstance(adaptiveUpdatePeriod,int):
        return "adaptiveUpdatePeriod=" + str(adaptiveUpdatePeriod) + inputFileDelim
    else:
        raise TypeError("The input specification, adaptiveUpdatePeriod, must be of type int.")

####################################################################################################################################

def greedyAdaptationCount(self,greedyAdaptationCount):
    if isinstance(greedyAdaptationCount,int):
        return "greedyAdaptationCount=" + str(greedyAdaptationCount) + inputFileDelim
    else:
        raise TypeError("The input specification, greedyAdaptationCount, must be of type int.")

####################################################################################################################################

def delayedRejectionCount(self,delayedRejectionCount):
    if isinstance(delayedRejectionCount,int):
        return "delayedRejectionCount=" + str(delayedRejectionCount) + inputFileDelim
    else:
        raise TypeError("The input specification, delayedRejectionCount, must be of type int.")

####################################################################################################################################

def burninAdaptationMeasure(self,burninAdaptationMeasure):
    if isinstance(burninAdaptationMeasure,(int,float)):
        return "burninAdaptationMeasure=" + str(burninAdaptationMeasure) + inputFileDelim
    else:
        raise TypeError("The input specification, burninAdaptationMeasure, must be of type float.")

####################################################################################################################################

def delayedRejectionScaleFactorVec(self,delayedRejectionScaleFactorVec):
    if isinstance(delayedRejectionScaleFactorVec,(list,tuple,np.ndarray)):
        return "delayedRejectionScaleFactorVec=" + str(np.array(list(delayedRejectionScaleFactorVec)).flatten()).strip('[]') + inputFileDelim
    else:
        raise TypeError("The input specification, delayedRejectionScaleFactorVec must be list, tuple, or numpy vector of ndim or less elements of type float.")

####################################################################################################################################

def helpme(specification = ""):
    """
    |
    Return help on the input ParaDRAM specification.
    If no input is provided, then the weblink to the 
    entire list of ParaDRAM specifications will be output.

        **Parameters**

            specification
                A string that can take be any of the simulation 
                specifications of a ParaDRAM object, such as, 
                ``"chainSize"``, ``"sampleSize"``, ...

        **Returns**

            None

    """
    import _paramonte as pm
    if isinstance(specification,str):
        if specification!="": specification = "#" + specification.lower()
    else:
        pm.abort( msg   = "The input argument to helpme() function must be a string whose value \n"
                        + "can be the name of one of the specifications of the ParaDRAM sampler.\n"
                        + "For example, to get information about the variable `chainSize`, try:\n\n"
                        + "    pmpd.spec.helpme(\"chainSize\")\n\n"
                        + "where ``pmpd`` is the name of the ParaDRAM sampler object.\n"
                        + "For more information and examples on the usage, visit:\n\n"
                        + "    " + pm.website.home.url
                , methodName = pm.names.paradram
                , marginTop = 1
                , marginBot = 1
                )

    print("Visit: " + pm.website.home.usage.paradram.specifications.url + specification)

####################################################################################################################################



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

#import typing as tp
from collections.abc import Iterable

class Struct: pass

####################################################################################################################################
#### nam2num
####################################################################################################################################

def nam2num ( nameList  # : tp.Union[ pd.core.indexes.base.Index , tp.List[str] ]
            , names     # : tp.List[str]
            ):
    nml = list(nameList)
    return [ nml.index(name) for name in names ]
    #if isinstance(nameList,list):
    #    return [ nameList.index(name) for name in names ]
    #else:
    #    
    #    return [ nameList.get_loc(name) for name in names ]

####################################################################################################################################
#### getColNamesIndex
####################################################################################################################################

def getColNamesIndex( dfColumns         # : pd.core.indexes.base.Index
                    , columns    = None # : tp.Optional[ tp.Union[ str, range , tp.List[int] , tp.List[str] ] ] = None
                    ):

    allColumnsNeeded = False
    if columns is None:
        allColumnsNeeded = True
    elif isinstance(columns,int):
        columns = [columns,]
    elif isinstance(columns,Iterable):
        if len(columns)==0: allColumnsNeeded = True
    else:
        raise Exception ( "\nThe input \"columns\" argument must be a list whose elements are all\n"
                        + "    1.   string-valued, each representing the name of the column from the dataframe, or,\n"
                        + "    2.   integer-valued, each representing the index of the column from the dataframe.\n"
                        + "You have entered:\n"
                        + str(columns)
                        )

    if allColumnsNeeded:
        colnames = dfColumns
        colindex = range(len(colnames))
    elif isinstance(columns, str):
        colnames = [columns]
        colindex = nam2num( dfColumns , [columns,] )
    elif all(isinstance(element, str) for element in columns):
        colnames = columns
        colindex = nam2num( dfColumns , columns )
    elif all(isinstance(element,int) for element in columns):
        colindex = columns
        colnames = dfColumns[colindex]
    #elif all(isinstance(element,float) for element in columns):
    #    colindex = columns
    #    colnames = None
    else:
        raise Exception ( "\nThe input argument 'columns' must be a list whose elements are all\n"
                        + "    1.   string-valued, each representing the name of the column from the dataframe, or,\n"
                        + "    2.   integer-valued, each representing the index of the column from the dataframe.\n"
                        + "You have entered:\n"
                        + str(columns)
                        )
    return colnames, colindex

####################################################################################################################################
#### getMaxLogFunc
####################################################################################################################################

def getMaxLogFunc( dataFrame # : pd.DataFrame
                 , column = "SampleLogFunc" # : tp.Optional[ str ] = "SampleLogFunc"
                 ):
    """

    Returns a structure with components containing the properties of the 
    input ``dataFrame``, corresponding to the mode (maximum) of the data 
    in the column of the dataFrame that is identified by ``column``. 

        **Parameters**

            dataFrame

                A Pandas dataframe containing the output sample data 
                from a ParaMonte simulation.

            column

                A string containing the name of the column of the input 
                ``dataFrame`` that contains values of the objective 
                function (or its logarithm).

        **Returns**

            maxLogFunc

                A structure with the following components:

                    idrow

                        The ID of the row of ``dataFrame`` corresponding 
                        to the mode of ``column``, 

                    value

                        The value of ``column`` at maximum, 

                    dfrow

                        The entire row of ``dataFrame`` corresponding to 
                        the mode of ``column``, 

                    state

                        The location (state) within the domain of the objective 
                        function where the maximum of ``column`` occurs.


    """

    _offset = dataFrame.columns.get_loc(column) + 1
    maxLogFunc = Struct()
    maxLogFunc.idrow = dataFrame[[column]].idxmax().values[0]
    maxLogFunc.value = dataFrame[[column]].iat[maxLogFunc.idrow,0]
    maxLogFunc.dfrow = dataFrame.iloc[maxLogFunc.idrow,:]
    maxLogFunc.state = dataFrame.iloc[maxLogFunc.idrow,_offset:]
    return maxLogFunc

####################################################################################################################################



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import typing as tp
import _pmutils as pmutils

Struct = pmutils.Struct
newline = pmutils.newline

####################################################################################################################################
#### Target class
####################################################################################################################################

class Target:
    """

    This is the Target class for generating instances
    of a target on the current active figure and axis.

        **Usage**

        First generate an object of this class by optionally
        passing the following parameters described below. Then call
        the ``make()`` method. The generated object is also callable 
        with the same input parameters as the object's constructor.

        **Parameters**

            values (optional)

                A pair (list, array, or tuple) of floats, 
                representing the (x,y) coordinates of the target.

            axes

                The axes object to which the target must be added.
                The default is ``None`` in which case the output of the 
                matplotlib's ``gca()`` function will be used to get the 
                current active axes.

        **Attributes**

            All of the parameters described above, except ``axes``.

                The input axes object (whether user-provided or automatically fetched)
                will be stored as a component of the object's attribute ``currentFig``.

            axvline (optional)

                A structure with two components:

                    enabled

                        A logical value, indicating whether an ``axvline`` 
                        will be added to the plot or not. If set to ``False``
                        no vertical target line will be added.

                    kws

                        A structure whose components will be passed to the 
                        matplotlib's ``axvline()`` function. For example:

                        .. code-block:: python
                            :linenos:

                            axvline.kws.linewidth = 0.5
                            axvline.kws.color = "orangered"
                            axvline.kws.linestyle = ":"
                            axvline.kws.ymin = 0.1
                            axvline.kws.ymax = 1.9

                        **NOTE**

                        If a desired property is missing in the 
                        structure, simply add the property and 
                        its value to the structure.

            axhline (optional)

                A structure with two components:

                    enabled

                        A logical value, indicating whether an ``axhline`` 
                        will be added to the plot or not. If set to ``False``
                        no horizontal target line will be added.

                    kws

                        A structure whose components will be passed to the 
                        matplotlib's ``axhline()`` function. For example:

                        .. code-block:: python
                            :linenos:

                            axhline.kws.linewidth = 0.5
                            axhline.kws.color = "orangered"
                            axhline.kws.linestyle = ":"
                            axhline.kws.xmin = 0.1
                            axhline.kws.xmax = 1.9

                        **NOTE**

                        If a desired property is missing in the 
                        structure, simply add the property and 
                        its value to the structure.

            scatter (optional)

                A structure with two components:

                    enabled

                        A logical value, indicating whether an ``scatter`` 
                        will be added to the plot or not. If set to ``False``
                        no target scatter circle will be added.

                    kws

                        A structure whose components will be passed to the 
                        matplotlib's ``scatter()`` function. For example:

                        .. code-block:: python

                            scatter.kws.s = 20
                            scatter.kws.color = "orangered"

                        NOTE: If a desired property is missing in the 
                        NOTE: structure, simply add the property and 
                        NOTE: its value to the structure.

            currentFig

                An structure which is initially ``None``, but upon making a plot, 
                is populated with attributes representing all outputs from 
                matplotlib function calls, with the following attributes:

                    axes

                        The output of matplotlib's ``gca()`` function or,
                        the input axes object by the user.

                    axhline

                        The output of matplotlib's ``axhline()`` function.

                    axvline

                        The output of matplotlib's ``axvline()`` function.

                    scatter

                        The output of matplotlib's ``scatter()`` function.

        **Returns**

            self

                An object of class ``Target``.

    ---------------------------------------------------------------------------
    """

    ################################################################################################################################
    #### __init__
    ################################################################################################################################

    def __init__( self
                , values : tp.Optional[ tp.Union[ tp.List[int] , tp.List[float] ] ] = [0,0]
                , axes = None #: tp.Optional[ plt.axes ] = None
                ):

        self.label = None
        self.value = values

        self.axvline = Struct()
        self.axvline.kws = Struct()
        self.axvline.enabled = True

        self.axhline = Struct()
        self.axhline.kws = Struct()
        self.axhline.enabled = True

        self.scatter = Struct()
        self.scatter.kws = Struct()
        self.scatter.enabled = True

        self.currentFig = Struct()
        self.currentFig.axes = axes
        if self.currentFig.axes is not None: self.currentFig.figure = self.currentFig.axes.get_figure()

        self._isdryrun = True
        self.make()
        self._isdryrun = False

    ################################################################################################################################
    #### __call__
    ################################################################################################################################

    def __call__( self
                , reself : tp.Optional[ bool ] = False
                , **kwargs
                ):
        """

        Call the ``make()`` method of the 
        current instance of the class.

            **Parameters**

                Any arguments that can be passed to the 
                ``make()`` method of the target object.

            **Returns**

                Any return value from the ``make()``
                method of the target object.

        """

        return self.make(reself, **kwargs)

    ################################################################################################################################
    #### make
    ################################################################################################################################

    def make( self
            , reself : tp.Optional[ bool ] = False
            , **kwargs
            ):
        """

        Add a target on an existing plot (the current active axes object)
        based on the ``values`` attribute of the target object.

            **Parameters**

                reself

                    A logical variable. If ``True``, an instance of 
                    the object will be returned upon exit to the calling 
                    routine. The default value is ``False``.

            **Returns**

                the object self if ``reself = True`` otherwise, ``None``.
                However, this method causes side-effects by manipulating
                the existing attributes of the object.

        """

        for key, val in kwargs.items():
            if hasattr(self,key):
                setattr(self, key, val)
            elif key=="axes":
                self.currentFig.axes = val
                self.currentFig.figure = self.currentFig.axes.get_figure()
            else:
                raise Exception ( "Unrecognized input '"+key+"' class attribute detected." + newline
                                + self._getDocString()
                                )

        # set what to plot

        if not isinstance(self.value, list): self.value = [self.value]

        ############################################################################################################################
        #### setup the vertical line
        ############################################################################################################################

        if self.axvline.enabled:
            if isinstance(self.axvline.kws,Struct):
                #if targetExists: self.axvline.kws.x = self.value[0]
                if "linewidth" not in vars(self.axvline.kws).keys(): self.axvline.kws.linewidth = 0.5
                if "linestyle" not in vars(self.axvline.kws).keys(): self.axvline.kws.linestyle = "-"
                if "zorder" not in vars(self.axvline.kws).keys(): self.axvline.kws.zorder = 1000
                if "color" not in vars(self.axvline.kws).keys(): self.axvline.kws.color = "orangered"
            else:
                raise Exception ( "The axvline.kws component of the current Target object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the axvline() function of the" + newline
                                + "matplotlib library." + newline
                                + self._getDocString()
                                )

        ############################################################################################################################
        #### setup the horizontal line
        ############################################################################################################################

        if self.axhline.enabled:
            if isinstance(self.axhline.kws,Struct):
                #if targetExists: self.axhline.kws.y = self.value[1]
                if "linewidth" not in vars(self.axhline.kws).keys(): self.axhline.kws.linewidth = 0.5
                if "linestyle" not in vars(self.axhline.kws).keys(): self.axhline.kws.linestyle = "-"
                if "color" not in vars(self.axhline.kws).keys(): self.axhline.kws.color = "orangered"
                if "zorder" not in vars(self.axhline.kws).keys(): self.axhline.kws.zorder = 1001
            else:
                raise Exception ( "The axhline.kws component of the current Target object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the axhline() function of the" + newline
                                + "matplotlib library." + newline
                                + self._getDocString()
                                )

        ############################################################################################################################
        #### setup the scatter line
        ############################################################################################################################

        if self.scatter.enabled:
            if isinstance(self.scatter.kws,Struct):
                #if targetExists: self.scatter.kws.x = self.value[0]
                #if targetExists: self.scatter.kws.y = self.value[1]
                if "s" not in vars(self.scatter.kws).keys(): self.scatter.kws.s = 20
                if "color" not in vars(self.scatter.kws).keys(): self.scatter.kws.color = "orangered"
                if "zorder" not in vars(self.scatter.kws).keys(): self.scatter.kws.zorder = 1002
            else:
                raise Exception ( "The scatter.kws component of the current Target object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the scatter() function of the" + newline
                                + "matplotlib library." + newline
                                + self._getDocString()
                                )

        ############################################################################################################################
        ############################################################################################################################
        if self._isdryrun: return
        ############################################################################################################################
        ############################################################################################################################

        from matplotlib import pyplot as plt

        # generate figure and axes if needed

        if self.currentFig.axes is None: self.currentFig.axes = plt.gca()
        self.currentFig.figure = self.currentFig.axes.get_figure()

        try:
            plt.sca(self.currentFig.axes)
        except:
            self.currentFig.axes = plt.gca()
            self.currentFig.figure = plt.gcf()
        #plt.figure(self.currentFig.figure.number)

        ############################################################################################################################
        #### add target
        ############################################################################################################################

        xlimCurrent = self.currentFig.axes.get_xlim()
        ylimCurrent = self.currentFig.axes.get_ylim()

        if self.axvline.enabled:

                #if "ymin" not in vars(self.axvline.kws).keys(): self.axvline.kws.ymin = ylimCurrent[0]
                #if "ymax" not in vars(self.axvline.kws).keys(): self.axvline.kws.ymax = ylimCurrent[1]
                self.currentFig.axvline = self.currentFig.axes.axvline  ( self.value[0]
                                                                        , **vars(self.axvline.kws)
                                                                        )

        if self.axhline.enabled:

                #if "xmin" not in vars(self.axhline.kws).keys(): self.axhline.kws.xmin = xlimCurrent[0]
                #if "xmax" not in vars(self.axhline.kws).keys(): self.axhline.kws.xmax = xlimCurrent[1]
                self.currentFig.axhline = self.currentFig.axes.axhline  ( self.value[1]
                                                                        , **vars(self.axhline.kws)
                                                                        )

        if self.scatter.enabled:

                self.currentFig.scatter = self.currentFig.axes.scatter  ( self.value[0]
                                                                        , self.value[1]
                                                                        , **vars(self.scatter.kws)
                                                                        )

        self.currentFig.axes.set_xlim(xlimCurrent)
        self.currentFig.axes.set_ylim(ylimCurrent)

        ############################################################################################################################

        if reself: return self

    ################################################################################################################################
    #### _getDocString
    ################################################################################################################################

    def _getDocString(self):
        docString   = newline \
                    + "Here is the help information on the Target class:" + newline \
                    + newline \
                    + self.__doc__
        return docString

    ################################################################################################################################



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np
import typing as tp
import weakref as wref
import _paramonte as pm
from paramonte.vis.Target import Target
from paramonte.vis._BasePlot import BasePlot

Struct = pm.Struct
newline = pm.newline

####################################################################################################################################
#### EllipsoidPlot class
####################################################################################################################################

class EllipsoidPlot(BasePlot):
    """

    This is the EllipsoidPlot class for generating instances of
    ellipsoid-evolution plots via matplotlib's builtin function 
    `plot()`. It generates a plot of 2D ellipsoids corresponding 
    to the input list of covariance/correlation matrices.

    **NOTE**

    This is a low-level ParaMonte class and is NOT meant
    to be directly instantiated by the users,
    although it is possible.

        **Parameters**

            matrix

                A 3D numpy array representing the covariance/correlation 
                matrices of the ellipsoids to plot. The size of the array 
                is ``(count, ndim, ndim)`` where ``count`` represents the 
                number of ellipsoids and ``ndim`` is in the number of 
                dimensions of the (hyper-)ellipsoids.

            plotType

                A string indicating the type of the plot to be made. 
                Possible values include: 

                    ``"covmat2"``
                    ``"covmat3"``
                    ``"cormat2"``
                    ``"cormat3"``

            methodName (optional)

                A string indicating the ParaMonte sampler that 
                is instantiating this EllipsoidPlot object. This 
                variable is solely used for progress reporting 
                and has no relevance to the plotting.
                The default value is ``"ParaMonte"``.

            reportEnabled (optional)

                A logical input parameter indicating whether the 
                ParaMonte automatic guidelines to the standard output 
                should be provided or not. The default value is ``True``.

            resetPlot (optional)

                An user-specified function that can be later called to 
                reset the settings of the instantiated Ellipsoid plot.
                If ``None`` is provided, then the `reset()`` of the 
                Ellipsoid object will be set to the default ``_reset()``
                function of the Ellipsoid object.
                The default value is ``None``.

        **Attributes**

            ndim

                An integer, representing the number of dimensions of the 
                domain of the objective function as inferred from the 
                contents of the dataFrame. It is the rank of the 
                input matrix.

            matrix

                The same as the corresponding input parameter ``matrix``
                to the ``EllipsoidPlot`` class constructor in the above.

            center

                A 2D numpy array representing the centers of the ellipsoids
                to plot. The size of the array is ``(count, ndim)`` where
                ``count`` represents the number of ellipsoids and ``ndim``
                is in the number of dimensions of the (hyper-)ellipsoids.

            zdata (optional, exists only in the case of 3D Ellipsoid plots)

                Optional property that, if provided, results in the creation 
                of 3D ellipsoid plots and its data represents the z-axis 
                values of the plot. It must be a numeric vector of the 
                same length as ``center`` and ``matrix``.

                The default value is the index of the covariance/correlation 
                matrices in the input data frame.

            dimensionPair

                A pair of indices (vector of length 2) whose value determine 
                the rows and columns from the covariance/correlation matrix 
                which will be plotted. 
                The default value is ``dimensionPair = [1,2]`` 
                (corresponding to Python indices of [0,1]).

                Example usage:

                    1.  ``dimensionPair = [1,2]``
                    2.  ``dimensionPair = [3,1]``

                **WARNING**

                In all cases, the indices must be distinct from 
                each other and less than ``ndim`` where ``ndim`` 
                is the rank of the covariance/correlation matrices.

            npoint

                The number of points used to represent the ellipsoids. 
                The higher the value of ``npoint``, the higher-resolution 
                the resulting ellipsoids will look like.

                The default value is ``100``.

            colorbar (exists only for plots that require colorbar)

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``colorbar()``
                        function of the matplotlib library should be made or not.
                        If a call is made, a new figure will be generated.
                        Otherwise, the current active figure will be used.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``colorbar()`` function of 
                        the matplotlib library.

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

                A colorbar will be added to a plot only if a color-mappings 
                is requested in the plot.

            colormap

                A structure with the following components:
                
                    enabled

                        A logical value indicating whether the ellipsoids 
                        must be color-mapped or not.

                    name

                        A string representing the name of the color mapping 
                        from the matplotlib library's ``cm`` class, to be 
                        applied to the plot. 

                    values

                        A numeric vector of the same length as the number 
                        of ellipsoids in ``matrix``, each element of which 
                        will determine the color of the corresponding 
                        ellipsoid in the plot.

                        If the values are set to ``None`` or an empty list, 
                        then a range of values corresponding to the count of 
                        ellipsoids will be used.

                    _rgba

                        The RGB vector corresponding to the values. This is 
                        not meant to be directly manipulated by the user.

                Example usage:

                    1.  ``colormap.enabled = True``
                    2.  ``colormap.name = "winter"``

            legend (may not exist for some types of plots)

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``legend()``
                        function of the matplotlib library should be made or not.
                        If a call is made, a new figure will be generated.
                        Otherwise, the current active figure will be used.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``legend()`` function.

                        Example usage:

                        .. code-block:: python

                            legend.kws.labels = ["Variable1", "Variable2"]

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

                A legend will be added to a plot only if no color-mappings 
                are requested in the plot.

            plot

                A structure with two components:

                    enabled

                        A logical value. If ``true``, 
                        the ellipsoids will be plotted.

                    kws
                        A structure whose components will be 
                        passed to the ``plot()`` function of 
                        the matplotlib library.

                        Example usage:

                            ``plot.kws.linewidth = 2``

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            title

                A structure with the following components:

                    enabled

                        A logical value. If ``true``, 
                        a title will be added to the plot.

                    label

                        A string representing the content of 
                        the title to be added to the plot.

                    kws

                        A structure whose components will be 
                        passed to the ``title()`` function 
                        of the matplotlib library.

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            target

                An object of class ``Target`` for adding target 
                values to the plots. For more information, see the 
                documentation of the ``Target()`` class.

        **Superclass Attributes**

            See the documentation for the ``BasePlot`` class.

        **Returns**

            An object of ``EllipsoidPlot`` class.

    ---------------------------------------------------------------------------
    """

    ################################################################################################################################
    #### __init__
    ################################################################################################################################

    def __init__( self
                , matrix        : tp.Union[ np.ndarray, tp.List[np.ndarray] ]
                , plotType      : str
                , methodName    : tp.Optional[ str ] = "ParaMonte"
                , reportEnabled : tp.Optional[ bool ] = True
                , resetPlot     = None
                ):

        self.matrix = matrix
        if isinstance(matrix, np.ndarray):
            self.numEllipsoid = np.shape(matrix)[0]
        #elif instance(matrix, list):
        #    self.numEllipsoid = len(matrix)
        else:
            pm.abort( msg   = "The input argument ``matrix`` must be a numpy 3D" + newline
                            + "array of matrices representing the covariance/correlation " + newline
                            + "matrices of the ellipsoids or. The first dimension of array" + newline
                            + "must count over the ellipsoids." + newline
                    , marginTop = 1
                    , marginBot = 1
                    , methodName = methodName
                    )
        import pandas as pd
        self._dfIndex = pd.DataFrame.from_dict({ "indices" : list(range( self.numEllipsoid )) })

        super().__init__( plotType = plotType
                        , dataFrame = self._dfIndex
                        , methodName = methodName
                        , reportEnabled = reportEnabled
                        , resetPlot = resetPlot
                        )

        self._reset()
        if resetPlot is None: self._resetPlot = self._reset
        self._progress.note()

    ################################################################################################################################
    #### _reset
    ################################################################################################################################

    def _reset(self):

        self.npoint = None
        self.center = None
        self.dimensionPair = None
        if self._type.is3d: self.zdata = None

        super()._reset()
        if not self._type.is3d: self.target = Target()

        self.plot = Struct()
        self.plot.enabled = True
        self.plot.kws = Struct()

        self.colormap = Struct()
        self.colormap.enabled = True
        self.colormap.values = None
        self.colormap.name = None

        self.colorbar = Struct()
        self.colorbar.enabled = True
        self.colorbar.kws = Struct()

        self.title = Struct()
        self.title.enabled = False
        self.title.label = ""
        self.title.kws = Struct()
        self.title.kws.loc = "center"

        self._isdryrun = True
        self.make()
        self._isdryrun = False

        self.legend.enabled = False
        self.legend._labels = None

    ################################################################################################################################
    #### __call__
    ################################################################################################################################

    def __call__( self
                , reself : tp.Optional[ bool ] = False
                , **kwargs
                ):
        """

        Call the ``make()`` method of the current 
        instance of the class.

            **Parameters**

                Any arguments that can be passed to the 
                ``make()`` method of the plot object.

            **Returns**

                Any return value from the ``make()`` 
                method of the plot object.

        """

        return self.make(reself, **kwargs)

    ################################################################################################################################
    #### make
    ################################################################################################################################

    def make( self
            , reself : tp.Optional[ bool ] = False
            , **kwargs
            ):
        """

        Generate an EllipsoidPlot from the input data the 
        EllipsoidPlot object.

            **Parameters**

                reself

                    A logical variable. If ``True``, an instance of the 
                    object will be returned upon exit to the calling 
                    routine. The default value is ``False``.

            **Returns**

                the object self if ``reself = True`` otherwise, ``None``.
                However, this method causes side-effects by manipulating
                the existing attributes of the object.

        """

        for key in kwargs.keys():
            if hasattr(self,key):
                setattr(self, key, kwargs[key])
            elif key=="dataFrame":
                setattr( self, "_dfref", wref.ref(kwargs[key]) )
            else:
                raise Exception ( "Unrecognized input '"+key+"' class attribute detected." + newline
                                + self._getDocString()
                                )

        if self.dimensionPair is None or (isinstance(self.dimensionPair,list) and len(self.dimensionPair)!=2): self.dimensionPair = [0,1]
        self.dimensionPair = np.int32(self.dimensionPair)
        if not isinstance(self.npoint, int): self.npoint = 100
        matrixShape = np.shape(self.matrix)
        self.numEllipsoid = matrixShape[0]
        self.ndim = matrixShape[1]

        # set what to plot

        ############################################################################################################################
        #### line plot properties
        ############################################################################################################################

        if isinstance(self.plot.kws,Struct):

            if "linewidth" in vars(self.plot.kws).keys():
                if self.plot.kws.linewidth==0: self.plot.kws.linewidth = 1
            else:
                self.plot.kws.linewidth = 1

            if "zorder" not in vars(self.plot.kws).keys(): self.plot.kws.zorder = 1

        else:

            raise Exception ( "The plot.kws component of the current EllipsoidPlot object must" + newline
                            + "be an object of class Struct(), essentially a structure with components" + newline
                            + "whose names are the input arguments to the plot() function of the" + newline
                            + "matplotlib library."
                            )

        ############################################################################################################################
        #### legend properties
        ############################################################################################################################

        if self.legend.enabled:
            if not isinstance(self.legend.kws,Struct):
                raise Exception ( "The legend.kws component of the current EllipsoidPlot object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the legend() function of the" + newline
                                + "matplotlib library."
                                )

        ############################################################################################################################
        #### figure properties
        ############################################################################################################################

        if self.figure.enabled:
            if isinstance(self.figure.kws, Struct):
                if "dpi" not in vars(self.figure.kws).keys(): self.figure.kws.dpi = 150
                if "facecolor" not in vars(self.figure.kws).keys(): self.figure.kws.facecolor = "w"
                if "edgecolor" not in vars(self.figure.kws).keys(): self.figure.kws.edgecolor = "w"
            else:
                raise Exception ( "The figure.kws component of the current EllipsoidPlot object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the figure() function of the" + newline
                                + "matplotlib library."
                                )

        ############################################################################################################################
        ############################################################################################################################
        if self._isdryrun: return
        ############################################################################################################################
        ############################################################################################################################

        from matplotlib import pyplot as plt

        ############################################################################################################################
        #### generate figure and axes if needed
        ############################################################################################################################

        self._constructBasePlot()

        ############################################################################################################################
        #### check data type
        ############################################################################################################################

        self._checkDataType()

        ############################################################################################################################
        #### check rows presence. This must be checked here, because it depends on the integrity of the in input dataFrame.
        ############################################################################################################################

        if self.rows is None: self.rows = range(len(self._dfref().index))

        ############################################################################################################################
        #### check the consistency of zdata
        ############################################################################################################################

        if self._type.is3d:
            if self.zdata is None:
                self.zdata = self._dfIndex.values
                self._zlabel = "Count"
            elif self.numEllipsoid != len(self.zdata):
                raise Exception ( "The lengths of the first dimension of the matrix and the zdata arrays must agree." + newline
                                + "This length is representative of the number of ellipsoids to plot."
                                )
            else:
                self._zlabel = None

        ############################################################################################################################
        #### set colormap data
        ############################################################################################################################

        if self.colormap.enabled:

            if self.colormap.values is None:
                self.colormap.values = self._dfIndex.values
                self.colorbar.label = "Count"
            else:
                self.colorbar.label = None

            self.colormap.values = np.double( self.colormap.values ) # ensure vectorized-indexing is possible

            if self.colormap.name is None or not isinstance(self.colormap.name, str):
                self.colormap.name = "winter"

            import matplotlib.colors as colors
            import matplotlib.cm as cmx

            colormapObject = plt.get_cmap(self.colormap.name)
            cNorm = colors.Normalize( vmin = np.min(self.colormap.values), vmax = np.max(self.colormap.values), clip = True )
            self.colorbar.kws.mappable = cmx.ScalarMappable( norm = cNorm, cmap = colormapObject )
            #self.colormap._rgba = colormapObject( self.colormap.values )
            self.colormap._rgba = np.zeros( ( len(self.rows) , 4) )
            for i, row in enumerate(self.rows):
                self.colormap._rgba[i,:] = self.colorbar.kws.mappable.to_rgba(self.colormap.values[row])

        ############################################################################################################################
        #### check the consistency of the lengths
        ############################################################################################################################

        if self.center is None: self.center = np.zeros( (self.numEllipsoid, self.ndim) )

        if self.numEllipsoid != np.shape(self.center)[0]: 
            raise Exception ( "The lengths of the first dimensions of the matrix and center vectors must agree." + newline
                            + "This length is representative of the number of ellipsoids to plot."
                            )

        ############################################################################################################################
        #### make ellipsoids
        ############################################################################################################################

        if self.plot.enabled: self.currentFig.plotList = []
        for i, row in enumerate(self.rows):

            if self._type.is3d: zdata = np.ones(self.npoint) * self.zdata[row]

            if self.colormap.enabled: self.plot.kws.color = self.colormap._rgba[i]

            fulMatrix = self.matrix[row,:,:]
            subMatrix = fulMatrix[:,self.dimensionPair][self.dimensionPair]

            # get ellipsoid boundary

            bcrd = self.getEllipsoidBoundary( covMat = subMatrix
                                            , meanVec = self.center[row,:]
                                            , npoint = self.npoint
                                            )

            if self.plot.enabled:

                if self._type.is3d:

                    self.currentFig.plotList.append( self.currentFig.axes.plot  ( bcrd[0,:]
                                                                                , bcrd[1,:]
                                                                                , zdata
                                                                                , **vars(self.plot.kws)
                                                                                ) )

                else:

                    self.currentFig.plotList.append( self.currentFig.axes.plot  ( bcrd[0,:]
                                                                                , bcrd[1,:]
                                                                                , **vars(self.plot.kws)
                                                                                ) )

        ############################################################################################################################
        #### add axes and title labels
        ############################################################################################################################

        # add axis labels

        self.currentFig.axes.set_xlabel("Dimension " + str(self.dimensionPair[0]+1))
        self.currentFig.axes.set_ylabel("Dimension " + str(self.dimensionPair[1]+1))
        if self._type.is3d and isinstance(self._zlabel, str): self.currentFig.axes.set_zlabel(self._zlabel)

        # set title properties

        if self.title.enabled:
            if not isinstance(self.title.label, str):
                raise Exception ("The title component of an EllipsoidPlot object must be a string or character vector.")
            else:
                plt.title( self.title.label, **vars(self.title.kws) )

        ############################################################################################################################
        #### add colorbar
        ############################################################################################################################

        if self.colormap.enabled and self.colorbar.enabled:
            #self.colorbar.kws.cax = self.currentFig.axes
            self.currentFig.colorbar = self.currentFig.figure.colorbar( **vars(self.colorbar.kws) )
            if isinstance(self.colorbar.label, str): self.currentFig.colorbar.set_label( label = self.colorbar.label )

        ############################################################################################################################
        #### finalize plot
        ############################################################################################################################

        self._finalizeBasePlot()

        ############################################################################################################################

        if reself: return self

    ################################################################################################################################
    #### getEllipsoidBoundary
    ################################################################################################################################

    def getEllipsoidBoundary( self
                            , covMat
                            , meanVec
                            , npoint = 50
                            ):
        """
        
        Return the coordinates of the boundary of an 
        ellipsoid represented by the input ``covMat``
        covariance matrix and the ``meanVec`` center.

            **Parameters**

                covMat

                    The representative covariance matrix 
                    of the ellipsoid.

                meanVec

                    A vector representing the center 
                    of the ellipsoid.

                npoint

                    The number of points with which the 
                    ellipsoid is represented. The more,
                    the higher the resolution will be.
                    The default value is 50.

            **Returns**

                A matrix of size ``(2, npoint)`` that contains 
                the boundary points of the generated ellipsoid 
                corresponding to the input matrix. 

        """

        from scipy import linalg as la

        independentVariable = np.linspace(0, 2*np.pi, npoint, endpoint = True)
        ap = np.zeros( (2, npoint) )
        ap[0,:] = np.cos(independentVariable)
        ap[1,:] = np.sin(independentVariable)
        eigenValues, eigenVecMat = la.eig(covMat)
        eigenValMat = np.eye(2)
        eigenValMat[0,0] = np.sqrt(eigenValues[0].real)
        eigenValMat[1,1] = np.sqrt(eigenValues[1].real)
        meanVecRepMat = np.ones( (2,npoint) )
        meanVecRepMat[0,:] = meanVec[0]
        meanVecRepMat[1,:] = meanVec[1]
        return np.matmul( np.matmul(eigenVecMat.real , eigenValMat) , ap ) + meanVecRepMat 

    ################################################################################################################################
    #### _getDocString
    ################################################################################################################################

    def _getDocString(self):
        docString   = newline \
                    + "Here is the help information on the Ellipsoid class:" + newline \
                    + newline \
                    + self.__doc__ \
                    + newline \
                    + "Here is the help information on the parent BasePlot class:" + newline \
                    + newline \
                    + super().__doc__
        return docString

    ################################################################################################################################
    #### helpme
    ################################################################################################################################

    def helpme(self, topic=None):
        """

        Print the documentation for the input string topic. 
        If the topic does not exist, the documentation for
        the object will be printed.
     
            **Parameters**
         
                topic (optional)

                A string containing the name of the object 
                for which help is needed.

            **Returns**
         
                None
         
            **Example**

                .. code-block:: python
                    :linenos:

                    helpme()
                    helpme("make")
                    helpme("helpme")
                    helpme("getLogLinSpace")

        """

        try:
            exec("print(self."+topic+".__doc__)")
        except:
            print(self._getDocString())
        return None

    ################################################################################################################################



#!/usr/bin/python
# Author:  Fatemeh Bagheri
# Contact: bagheri.fateme@gmail.com
####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################
"""

This module contains the visualization tools of the ParaMonte Python library.

Normally, the user is not supposed to access this module or its contents 
directly. However, it is made accessible for the purpose of visualizing 
data generated by packages other than ParaMonte. In either case, the 
structure of the input datasets are expected to follow the rules and 
conventions specified by each tool. 

-------------------------------------------------------------------------------
"""

####################################################################################################################################



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np
from scipy.fft import dctn, idctn
from scipy.optimize import brentq

####################################################################################################################################
#### kde2d
####################################################################################################################################

def kde2d(x, y, n=256, limits=None):
    """

    Return the 2d density map from discrete observations via 
    2-dimensional diffusion Kernel density estimation.

    First the input data is binned. After binning, the function 
    determines the optimal bandwidth according to the diffusion-based 
    method. It then smooths the binned data over the grid using a 
    Gaussian kernel with a standard deviation corresponding to 
    that bandwidth.

        This module is based on the KDE estimation method of 

            Z. I. Botev, J. F. Grotowski, D. P. Kroese: 
            Kernel density estimation via diffusion. 
            Annals of Statistics 38 (2010), no. 5, 2916--2957. 
            doi:10.1214/10-AOS799

        and 

            John Hennig 
            DOI: 10.5281/zenodo.3830437 

        **Parameters**

            x

                A lists of array of numbers that represent discrete 
                observations of a random variable with two coordinate 
                components. The observations are binned on a grid of 
                n*n points, where ``n`` must be a power of 2 or will 
                be coerced to the next one. If ``x`` and ``y`` are not 
                the same length, the algorithm will raise a ``ValueError``.

            y

                A lists of array of numbers that represent discrete 
                observations of a random variable with two coordinate 
                components. The observations are binned on a grid of 
                n*n points, where ``n`` must be a power of 2 or will 
                be coerced to the next one. If ``x`` and ``y`` are not 
                the same length, the algorithm will raise a ``ValueError``.

            n (optional)

                The number of grid points. It must be a power of 2. 
                Otherwise, it will be coerced to the next power of two.
                The default is 256.

            limits (optional)

                Data ``limits`` specified as a tuple of tuples denoting
                ``((xmin, xmax), (ymin, ymax))``. If any of the values 
                are ``None``, they will be inferred from the data. Each 
                tuple, or even both of them, may also be replaced by a 
                single value denoting the upper bound of a range 
                centered at zero. The default is ``None``.

        **Returns**

            A tuple whose elements are the following:

                density

                    The density map of the data.

                grid

                    The grid at which the density is computed.

                bandwidth

                    The optimal values (per axis) that the algorithm has 
                    determined. If the algorithm does not converge, it 
                    will raise a ``ValueError``.

    ---------------------------------------------------------------------------
    """

    # Convert to arrays in case lists are passed in.

    x = np.array(x)
    y = np.array(y)

    # Make sure numbers of data points are consistent.

    N = len(x)
    if len(y) != N:
        raise ValueError('x and y must have the same length.')

    # Round up number of bins to next power of two.

    n = int(2**np.ceil(np.log2(n)))

    # Determine missing data limits.

    if limits is None:

        xmin = xmax = ymin = ymax = None

    elif isinstance(limits, tuple):

        (xlimits, ylimits) = limits
        if xlimits is None:
            xmin = xmax = None
        elif isinstance(xlimits, tuple):
            (xmin, xmax) = xlimits
        else:
            xmin = -xlimits
            xmax = +xlimits
        if ylimits is None:
            ymin = ymax = None
        elif isinstance(ylimits, tuple):
            (ymin, ymax) = ylimits
        else:
            ymin = -ylimits
            ymax = +ylimits
    else:

        xmin = -limits
        xmax = +limits
        ymin = -limits
        ymax = +limits

    if None in (xmin, xmax):
        delta = x.max() - x.min()
        if xmin is None:
            xmin = x.min() - delta/4
        if xmax is None:
            xmax = x.max() + delta/4

    if None in (ymin, ymax):
        delta = y.max() - y.min()
        if ymin is None:
            ymin = y.min() - delta/4
        if ymax is None:
            ymax = y.max() + delta/4

    deltax = xmax - xmin
    deltay = ymax - ymin

    # Bin samples on regular grid.

    (binned, xedges, yedges) = np.histogram2d(x, y, bins=n, range=((xmin, xmax), (ymin, ymax)))
    grid = (xedges[:-1], yedges[:-1])

    # Compute discrete cosine transform. Adjust first component.

    transformed = dctn(binned/N)
    transformed[0, :] /= 2
    transformed[:, 0] /= 2

    # Pre-compute squared indices and transform components before solver loop.

    k  = np.arange(n, dtype="float") # float avoids integer overflow.
    k2 = k**2
    a2 = transformed**2

    # Define internal functions to be solved iteratively.

    def γ(t):
        sigma = psi(0, 2, t) + psi(2, 0, t) + 2*psi(1, 1, t)
        γ = (2*np.pi*N*sigma)**(-1/3)
        return (t - γ) / γ

    def psi(i, j, t):
        if i + j <= 4:
            sigma  = abs(psi(i+1, j, t) + psi(i, j+1, t))
            C  = (1 + 1/2**(i+j+1)) / 3
            pii = np.product(np.arange(1, 2*i, 2))
            pij = np.product(np.arange(1, 2*j, 2))
            t  = (C*pii*pij / (np.pi*N*sigma)) ** (1/(2+i+j))
        w = 0.5 * np.ones(n)
        w[0] = 1
        w = w * np.exp(-np.pi**2 * k2*t)
        wx = w * k2**i
        wy = w * k2**j
        return (-1)**(i+j) * np.pi**(2*(i+j)) * wy @ a2 @ wx

    # Solve for optimal diffusion time t*.

    try:
        ts = brentq(lambda t: t - γ(t), 0, 0.1)
    except ValueError:
        raise ValueError('Bandwidth optimization did not converge.') from None

    # Calculate diffusion times along x- and y-axis.

    psi02 = psi(0, 2, ts)
    psi20 = psi(2, 0, ts)
    psi11 = psi(1, 1, ts)
    tx1 = (psi02**(3/4) / (4*np.pi*N*psi20**(3/4) * (psi11 + np.sqrt(psi02*psi20))) )**(1/3)
    tx2 = (psi20**(3/4) / (4*np.pi*N*psi02**(3/4) * (psi11 + np.sqrt(psi02*psi20))) )**(1/3)

    # Note:
    # The above uses the nomenclature from the paper. In the Matlab
    # reference, tx1 is called t_y, while tx2 is t_x. This is a curious
    # change in notation. It may be related to the fact that image
    # coordinates are typically in (y,x) index order, whereas matrices,
    # such as the binned histogram (in Matlab as much as in Python),
    # are in (x,y) order. The Matlab code eventually does return
    # image-like index order, though it never explicitly transposes
    # the density matrix. That is implicitly handled by its custom
    # implementation of the inverse transformation (idct2d), which
    # only employs one matrix transposition, not two as its forward
    # counterpart (dct2d).

    # Apply Gaussian filter with optimized kernel.

    smoothed = transformed * np.outer(np.exp(-np.pi**2 * k2 * tx2/2), np.exp(-np.pi**2 * k2 * tx1/2))

    # Reverse transformation.

    smoothed[0, :] *= 2
    smoothed[:, 0] *= 2
    inverse = idctn(smoothed)

    # Normalize density.

    density = np.transpose(inverse) * n/deltax * n/deltay

    # Determine bandwidth from diffusion times.

    bandwidth = np.array([np.sqrt(tx2)*deltax, np.sqrt(tx1)*deltay])

    # Return results.

    return (density, grid, bandwidth)



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import numpy as np
import _paramonte as pm
import _CorCovMat as ccm
from _OutputFileContents import OutputFileContents
from paramonte.vis.LineScatterPlot import LineScatterPlot
from paramonte.vis.EllipsoidPlot import EllipsoidPlot

Struct = pm.Struct
newline = pm.newline

# amazingly strange. If Struct is taken from outside this module,
# it will automatically add: isParaDRAM, isParaNest, isParaTemp attributes
# print("\nself.contents.isParaDRAM: {}\n".format(self.contents.isParaDRAM))
# issue resolved: self._method = pm.Struct in _ParaMonteSampler (without parentheses, it does not instantiate).
# class Struct: pass

####################################################################################################################################
#### RestartFileContents
####################################################################################################################################

class RestartFileContents(OutputFileContents):
    """

    This is the **RestartFileContents** class for generating instances
    of the ParaMonte restart output contents. This class is NOT meant to
    be directly accessed by the ParaMonte library users. It is internally
    used by the ParaMonte library to parse the contents of the output restart
    files generated by the ParaMonte sampler routines. For example, the ParaDRAM
    sampler class makes calls to this class via its ``readRestart()`` method to
    return a list of objects of class ``RestartFileContents``.

        **Parameters**

            file

                The full path to the file containing the sample/chain.

            methodName

                A string representing the name of the ParaMonte sampler used
                to call the constructor of the ``RestartFileContents`` class.

            reportEnabled

                A logical input parameter indicating whether the ParaMonte
                automatic guidelines to the standard output should be provided 
                or not. The default value is ``True``.

        **Attributes**

            A dynamic set of attributes that are directly parsed from the file.

        **Returns**

            restartFileContents

                An object of class ``RestartFileContents``.

    ----------------------------------------------------------------------
    """

    ################################################################################################################################
    #### __init__
    ################################################################################################################################

    def __init__( self
                , file
                , methodName
                , reportEnabled
                ):

        super().__init__(file, methodName, reportEnabled)
        # superclass sets the following
        # self.file = file
        # self._methodName = methodName
        # self._reportEnabled = reportEnabled

        self.ndim = None
        self.count = None
        self.contents = None
        self._fileType = "restart"
        self._contents = []
        self._lineList = []
        self._lineListLen = []
        self.propNameList = []

        if self._methodName == pm.names.paradram:
            self._readRestartParaDRAM()
        else:
            pm.abort( msg = "Internal error occurred. Unrecognized methodName in the class \n"
                            "constructor of RestartFileContents: " + self.methodName
                    , marginTop = 1
                    , marginBot = 1
                    , methodName = "ParaMonte"
                    )

    ################################################################################################################################
    #### _readRestartParaDRAM
    ################################################################################################################################

    def _readRestartParaDRAM(self):

        # ParaDRAM field names

        self.propNameList = [ "meanAcceptanceRateSinceStart"
                            , "sampleSize"
                            , "logSqrtDeterminant"
                            , "adaptiveScaleFactorSquared"
                            , "meanVec"
                            , "covMat"
                            , "corMat"
                            ]

        ############################################################################################################################
        #### data
        ############################################################################################################################

        with open(self.file,"r") as fid: self._contents = fid.read().replace(pm.creturn,"")
        self._lineList = self._contents.split(pm.newline)
        self._lineListLen = len(self._lineList)

        # find count of updates

        self.count = self._contents.count(self.propNameList[0])

        # find ndim via meanVec entry: self.propNameList[4]

        rowOffset = 0;
        while self.propNameList[4] not in self._lineList[rowOffset]:
            rowOffset = rowOffset + 1
            if rowOffset > self._lineListLen: self._reportCorruptFile()
        rowOffset = rowOffset + 1 # the first numeric value of the meanVec

        self.ndim = 0
        while pm.utils.isNumericString( self._lineList[rowOffset+self.ndim] ): self.ndim += 1
        if self.ndim==0: self._reportCorruptFile()

        # parse the restart file contents

        fieldNamesDict =    { self.propNameList[0] : np.zeros(self.count)
                            , self.propNameList[1] : np.zeros(self.count)
                            , self.propNameList[2] : np.zeros(self.count)
                            , self.propNameList[3] : np.zeros(self.count)
                            , self.propNameList[4] : np.zeros((self.count,self.ndim))
                            , self.propNameList[5] : np.zeros((self.count,self.ndim,self.ndim))
                            , self.propNameList[6] : np.zeros((self.count,self.ndim,self.ndim))
                            }

        skip = 10 + (self.ndim * (self.ndim + 3)) // 2
        progressFraction = np.floor( self.count / 20 )
        for icount in range(self.count):

            if icount%progressFraction==0: self._progress.updateBar(icount/self.count-1)

            istart = icount * skip + 1

            rowOffset = 0
            fieldNamesDict[self.propNameList[0]][icount] = np.double( self._lineList[istart+rowOffset] )

            rowOffset = 2
            fieldNamesDict[self.propNameList[1]][icount] = np.double( self._lineList[istart+rowOffset] )

            rowOffset = 4
            fieldNamesDict[self.propNameList[2]][icount] = np.double( self._lineList[istart+rowOffset] )

            rowOffset = 6
            fieldNamesDict[self.propNameList[3]][icount] = np.double( self._lineList[istart+rowOffset] )

            rowOffset = 8
            iend = istart + rowOffset + self.ndim
            fieldNamesDict[self.propNameList[4]][icount,:] = np.double( self._lineList[istart+rowOffset:iend] )

            iend += 1 # the first numeric element of the covariance matrix
            for i in range(self.ndim): # covmat
                iPlusOne = i + 1
                istart = iend
                iend = istart + iPlusOne
                #print("\n{}".format(np.double( self._lineList[istart:iend]) ))
                fieldNamesDict[self.propNameList[5]][icount,i,0:iPlusOne] = np.double( self._lineList[istart:iend] )
                fieldNamesDict[self.propNameList[5]][icount,0:iPlusOne,i] = fieldNamesDict[self.propNameList[5]][icount,i,0:iPlusOne]

            fieldNamesDict[self.propNameList[6]][icount,:,:] = ccm.getCorFromCov( fieldNamesDict[self.propNameList[5]][icount,:,:] )

            #print("\n") # xxx
            #print(self.ndim) # xxx
            #print(self.count) # xxx
            #print(self.propNameList[0]+"\n"+str(fieldNamesDict[self.propNameList[0]][icount])) # xxx
            #print(self.propNameList[1]+"\n"+str(fieldNamesDict[self.propNameList[1]][icount])) # xxx
            #print(self.propNameList[2]+"\n"+str(fieldNamesDict[self.propNameList[2]][icount])) # xxx
            #print(self.propNameList[3]+"\n"+str(fieldNamesDict[self.propNameList[3]][icount])) # xxx
            #print(self.propNameList[4]+"\n"+str(fieldNamesDict[self.propNameList[4]][icount])) # xxx
            #print(self.propNameList[5]+"\n"+str(fieldNamesDict[self.propNameList[5]][icount,:,:])) # xxx
            #print(ccm.getCorFromCov( fieldNamesDict[self.propNameList[6]][icount,:,:] )) # xxx
            #break # xxx

        self._progress.updateBar(1)

        self.contents = Struct()
        for fieldName in self.propNameList: setattr(self.contents, fieldName, fieldNamesDict[fieldName])
        _ = fieldNamesDict.pop(self.propNameList[4])
        _ = fieldNamesDict.pop(self.propNameList[5])
        _ = fieldNamesDict.pop(self.propNameList[6]) # get rid of meanVec, covMat, corMat
        import pandas as pd
        self.df = pd.DataFrame.from_dict(fieldNamesDict)

        self._progress.note()

        ############################################################################################################################
        #### graphics
        ############################################################################################################################

        self._plotTypeList =    [ "line"
                                , "scatter"
                                , "lineScatter"
                                ]

        if self.ndim>1: self._plotTypeList +=   [ "covmat2"
                                                , "covmat3"
                                                , "cormat2"
                                                , "cormat3"
                                                ]

        self._progress.note( msg = "adding the graphics tools... ", end = newline, pre = True )
        self.plot = Struct()
        self._resetPlot(resetType="hard")
        self._progress.note()

        self.plot.reset = self._resetPlot

    ################################################################################################################################
    #### _reportCorruptFile
    ################################################################################################################################

    def _reportCorruptFile(self):
        pm.abort( msg = "The structure of the file: " + newline
                        + newline
                        + "    \"" + self.file + "\"" + newline
                        + newline
                        + "does not match a " + self._methodName + " " + self._fileType + " file." + newline
                        + "The contents of the file may have been compromised." + newline
                        + "Verify the integrity of the contents of this file before attempting to reread it."
                , marginTop = 2
                , marginBot = 1
                , methodName = self._methodName
                )

    ################################################################################################################################
    #### _resetPlot
    ################################################################################################################################

    def _resetPlot  ( self
                    , resetType = "soft"
                    , plotNames = "all"
                    ):
        """

        Reset the properties of the plot to the original default settings.
        Use this method when you change many attributes of the plot and
        you want to clean up and go back to the default settings.

            **Parameters**

                resetType (optional)

                    An optional string with possible value of ``"hard"``.
                    If provided, the plot object will be regenerated from scratch.
                    This includes reading the original data frame again and resetting
                    everything. If not provided, then only the plot settings will be
                    reset without reseting the dataFrame.

                plotNames (optional)

                    An optional string value or list of string values representing 
                    the names of plots to reset. If no value is provided, 
                    then all plots will be reset.

            **Returns**

                None

            **Example**

                .. code-block:: python

                    reset("hard")                       # regenerate all plots from scratch
                    reset("hard","covmat2")             # regenerate covmat2 plot from scratch
                    reset("hard",["covmat2","covmat3"]) # regenerate covmat2 & covmat3 plots

        """

        requestedPlotTypeList = []
        if isinstance(plotNames, str):
            plotTypeLower = plotNames.lower()
            if plotTypeLower=="all":
                requestedPlotTypeList = self._plotTypeList
            elif plotNames in self._plotTypeList:
                requestedPlotTypeList = [plotNames]
            else:
                self._reportWrongPlotName(plotNames)
        elif isinstance(plotNames, list):
            for plotName in plotNames:
                if plotName not in self._plotTypeList: self._reportWrongPlotName(plotName)
        else:
            self._reportWrongPlotName("a none-string none-list object.")

        resetTypeIsHard = None
        if isinstance(resetType, str):
            resetTypeIsHard = resetType.lower()=="hard"
        else:
            resetTypeIsHard = None
            pm.abort( msg   = "The input argument resetType must be a string representing" + newline
                            + "the type of the reset to be performed on the plots." + newline
                            + "A list of possible plots includes: \"hard\", \"soft\"" + newline
                            + "Here is the help for the ``reset()`` method: " + newline
                            + newline
                            + self._resetPlot.__doc__
                    , marginTop = 1
                    , marginBot = 1
                    , methodName = self._methodName
                    )

        ############################################################################################################################
        #### reset plots
        ############################################################################################################################

        for requestedPlotType in requestedPlotTypeList:

            plotObject = None
            requestedPlotTypeLower = requestedPlotType.lower()

           #is3d        = "3"           in requestedPlotTypeLower
            isLine      = "line"        in requestedPlotTypeLower
            isScatter   = "scatter"     in requestedPlotTypeLower
            isCovMat    = "covmat"      in requestedPlotTypeLower
            isCorMat    = "cormat"      in requestedPlotTypeLower

            if not resetTypeIsHard:
                plotComponent = getattr(self, "plot")
                plotObject = getattr(plotComponent, requestedPlotType)
                plotObject._reset()

            ########################################################################################################################
            #### reset line / scatter
            ########################################################################################################################

            if isLine or isScatter:

                if resetTypeIsHard:
                    plotObject = LineScatterPlot( plotType = requestedPlotType
                                                , dataFrame = self.df
                                                , methodName = self._methodName
                                                , reportEnabled = self._reportEnabled
                                                , resetPlot = self._resetPlot
                                                )

                if self._methodName == pm.names.paradram: plotObject.ycolumns = self.propNameList[0]
                plotObject.ccolumns = []
                plotObject.colorbar.kws.extend = "neither"
                plotObject.colorbar.kws.orientation = "vertical"
                plotObject.colorbar.kws.spacing = "uniform"

                if isLine:
                    if isScatter:
                        plotObject.lineCollection.enabled = False
                        plotObject.plot.enabled = True
                        plotObject.plot.kws.alpha = 0.2
                        plotObject.plot.kws.color = "grey"
                        plotObject.plot.kws.linewidth = 0.75
                    else:
                        plotObject.lineCollection.enabled = True
                        plotObject.plot.enabled = False

                setattr(self.plot, requestedPlotType, plotObject)

            ########################################################################################################################
            #### reset covmat / cormat
            ########################################################################################################################

            if isCovMat or isCorMat:

                if self._methodName == pm.names.paradram:

                    matrix = None
                    if isCovMat: matrix = self.contents.covMat
                    if isCorMat: matrix = self.contents.corMat

                    if resetTypeIsHard:
                        plotObject = EllipsoidPlot  ( matrix = matrix
                                                    , plotType = requestedPlotType
                                                    , methodName = self._methodName
                                                    , reportEnabled = self._reportEnabled
                                                    , resetPlot = self._resetPlot
                                                    )

                    plotObject.rows = plotObject.getLogLinSpace()
                    plotObject.center = self.contents.meanVec
                    plotObject.title.enabled = True
                    matrixType = "covariance" if isCovMat else "correlation"
                    plotObject.title.label = "Evolution of the " + matrixType + " matrices of the proposal distribution"

                    plotObject.colorbar.kws.extend = "neither"
                    plotObject.colorbar.kws.orientation = "vertical"
                    plotObject.colorbar.kws.spacing = "uniform"

                    setattr(self.plot, requestedPlotType, plotObject)

    ################################################################################################################################



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import _message as err
import numpy as np
import time
import sys
import os

class Struct: pass

newline = chr(10)
creturn = chr(13)

####################################################################################################################################
#### Frozen Struct
####################################################################################################################################

class FrozenClass(object):
    __isfrozen = False
    def __setattr__(self, key, value):
        if self.__isfrozen and not hasattr(self, key):
            raise TypeError ( "\n{} is a frozen class.\n".format(self)
                            + "The requested attribute '{}' does not exist in the object.\n".format(key)
                            + "You cannot add new attributes to an object of frozen class."
                            )
        object.__setattr__(self, key, value)

    def _freeze(self):
        self.__isfrozen = True

####################################################################################################################################
#### isNumericString
####################################################################################################################################

def isNumericString(string):
    try:
        np.double(string)
        return True
    except ValueError:
        return False

####################################################################################################################################
#### object size estimation
####################################################################################################################################

def getSize(obj, seen=None):
    """Recursively finds size of objects"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([getSize(v, seen) for v in obj.values()])
        size += sum([getSize(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += getSize(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([getSize(i, seen) for i in obj])
    return size

####################################################################################################################################
#### check application installation status
####################################################################################################################################

def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None

####################################################################################################################################
#### convert to list
####################################################################################################################################

def getList(x):
    """
    Convert any input object to list.
    """
    if isinstance(x, list):
        return x
    elif isinstance(x, str):
        return [x]
    try:
        return list(x)
    except TypeError:
        return [x]

####################################################################################################################################
#### delFile
####################################################################################################################################

def delFile(file, desc="the requested file", methodName=""):
    try:
        import os
        os.remove(file) 
    except:
        err.warn( msg   = newline
                        + "Failed to delete " + desc + " from the local disk. File may be protected. You can try manually deleting it: " + newline
                        + newline
                        + "    " + file + newline
                        + newline
                , methodName = methodName
                , marginTop = 1
                , marginBot = 1
                )
    return None

####################################################################################################################################
#### get file list
####################################################################################################################################

def getFileList(file, fileSuffix, methodName, reportEnabled = True):

    FileList = []
    iswebfile = False
    fullSuffix = "_" + fileSuffix + ".txt"

    if os.path.isfile(file):

        # check if the input path is a full path to a file

        FileList.append(file)
        pattern = file
        if fullSuffix != file[-len(fullSuffix):]:
            err.warn( msg   = "The name of the input file: \n\n"
                            + "    " + file + "\n\n"
                            + "does not end with the expected suffix '" + fullSuffix + "' for a " + fileSuffix + " file type.\n"
                    , methodName = methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

    else:

        import glob

        pattern = "" # not really needed, but just in case...
        for i in [1,2]:

            if i==1:

                #### first search for files matching the input pattern, then for directory

                pattern = file if "*" in file else file + "*" # + fullSuffix

            elif i==2:

                #### then search for file as directory

                if os.path.isdir(file):

                    pattern = os.path.join(file, "*" + fullSuffix)

                    if reportEnabled:
                        err.warn( msg   = "file='" + file + "' points to a directory.\n"
                                        + "Now searching inside the folder for a " + fileSuffix + " file..."
                                , methodName = methodName
                                , marginTop = 1
                                , marginBot = 1
                                )

            #### now search for pattern

            for filePath in glob.glob(pattern):
                if filePath.endswith(fullSuffix):
                    FileList.append(filePath)

            if len(FileList)>0: break

        if not pattern.endswith(fullSuffix): pattern += fullSuffix

        if len(FileList)==0:

            #### one last try, search the web

            try:
                import urllib.request
                #filePath = getRandomFilePrefix(prefix = methodName + "_" + fileSuffix)
                localFilePath, headers = urllib.request.urlretrieve(url = file)
                FileList.append(localFilePath)
                iswebfile = True
            except:
                err.abort   ( msg   = "Failed to detect any " + fileSuffix + " files with the requested pattern: \n\n"
                                    + "    " + pattern + "\n\n"
                                    + "Provide a string, as the value of the input argument ``file``, that either \n\n"
                                    + "    - points to one or more " + fileSuffix + " files, or, \n"
                                    + "    - represents the unique name of a ParaMonte simulation. \n"
                                    + "      This unique-name is the common prefix in the names of \n"
                                    + "      the output files of a ParaMonte simulation.\n\n"
                                    + "Most importantly, ensure the requested file is in ASCII format.\n"
                                    + "The binary-format chain or restart output files cannot be parsed.\n"
                                    + "You can request ASCII-format output files by setting the\n"
                                    + "appropriate simulation specifications of the " + methodName + " sampler,\n\n"
                                    + "    spec.restartFileFormat = \"ascii\"\n"
                                    + "    spec.chainFileFormat = \"ascii\""
                            , methodName = methodName
                            , marginTop = 1
                            , marginBot = 1
                            )

        elif reportEnabled:

            err.note( msg = str(len(FileList)) + ' files detected matching the pattern: "' + pattern + '"'
                    , methodName = methodName
                    , marginTop = 1
                    , marginBot = 1
                    )

    return FileList, iswebfile

####################################################################################################################################
#### Timer class
####################################################################################################################################

class Timer:

    def __init__(self):
        self.start = 0
        self.total = 0
        self.delta = 0
        self.last = 0
        self.tic()

    def tic(self):
        self.start = time.time()
        self.total = self.start
        self.last = self.total
        self.delta = self.total - self.last

    def toc(self):
        self.last = self.total
        self.total = time.time()
        self.delta = self.total - self.last

####################################################################################################################################
#### Progress report
####################################################################################################################################

class Progress:

    def __init__( self
                , msg = None
                , methodName = ""
                , reportEnabled = True
                , end = "\n"
                ):
        self._methodName = methodName
        self._reportEnabled = reportEnabled
        self._oldFraction = 0.0
        self._clockCounter = None
        if self._reportEnabled and msg is not None:
            err.note( msg = msg
                    , methodName = self._methodName
                    , marginTop = 0
                    , marginBot = 0
                    , end = end
                    )
        self.timer = Timer()

    ################################################################################################################################

    def note( self
            , msg = None
            , end = "\n"
            , pre = False
            ):
        self.timer.toc()
        if self._reportEnabled:
            if msg is None: msg = "done in " + str(np.round(self.timer.delta,decimals=6)) + " seconds."
            if pre:
                err.note( msg = msg
                        , methodName = self._methodName
                        , marginTop = 0
                        , marginBot = 0
                        , end = end
                        )
            else:
                print( msg, end = end )
        self.timer.toc()

    ################################################################################################################################

    # dynamic progress bar
    def updateBar   ( self
                    , fraction
                    , progressFraction = 0.05
                    ):
        if self._reportEnabled:
            if fraction<1:
                end = ""
            else:
                end = "\n"
            if fraction > self._oldFraction + progressFraction:
                self._oldFraction = fraction
                print( ".", end=end);

    ################################################################################################################################

    # dynamic clock tic
    def updateClock ( self
                    , fraction
                    ):
        if self._reportEnabled:
            chars = "|/-\\"
            if self._clockCounter is None:
                self._clockCounter = 0
                backspace = "\r"
            else:
                self._clockCounter = self._clockCounter%3 + 1
                backspace = 5*"\r"
            if fraction<1:
                clockTick = chars[self._clockCounter]; 
                end = ""
            else:
                self._clockCounter = None;
                clockTick = chars[0];
                end = "\n"
            print( backspace + clockTick + "{:3.0f}%".format(100*fraction), end=end);

####################################################################################################################################
#### getRandomFilePrefix
####################################################################################################################################

def getRandomFilePrefix(prefix = ""):
    from datetime import datetime as dt
    now = dt.now()
    return  prefix \
            + "{:04d}".format(now.year) + "{:02d}".format(now.month) + "{:02d}".format(now.day) + "_"  \
            + "{:02d}".format(now.hour) + "{:02d}".format(now.minute) + "{:02d}".format(now.second) + "_" \
            + "{:03d}".format(round(now.microsecond/1000))

####################################################################################################################################
#### getLogIntSpace
####################################################################################################################################

def getLogIntSpace(base, logskip, lowerLim, upperLim):
    """
    return a set of unique integer spacings linearly-spaced in the 
    logarithmic scale in the input given base, between the input limits.
    """
    if base<=1: raise ValueError("The input argument \"base\" must be a positive real number. You have entered: " + str(base))
    if logskip<=0: raise ValueError("The input argument \"logskip\" must be a positive real number. You have entered: " + str(logskip))
    if lowerLim<=0: raise ValueError("The input argument \"lowerLim\" must be a positive real number. You have entered: " + str(lowerLim))
    if upperLim<=lowerLim: raise ValueError ( "The input argument \"upperLim\" must be a positive real number larger than "
                                            + "the input argument \"lowerLim\". You have entered: " + str(upperLim))
    return np.unique( np.int32( base**( np.arange( np.log(lowerLim)/np.log(base), np.log(upperLim)/np.log(base), logskip) ) ) );

####################################################################################################################################



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import typing as tp
import weakref as wref

import _paramonte as pm
import _pmutils as pmutils
from paramonte.vis._BasePlot import BasePlot

Struct = pmutils.Struct
newline = pmutils.newline

####################################################################################################################################
#### HeatMapPlot class
####################################################################################################################################

class HeatMapPlot(BasePlot):
    """

    This is the HeatMapPlot class for generating instances of 
    heatmap figures based on the seaborn library's ``heatmap()``.

        **Usage**

            First generate an object of this class by optionally
            passing the following parameters described below. 
            Then call the ``make()`` method.

        **Parameters**

            plotType

                A string indicating the name of the plot to be constructed.

            dataFrame (optional)

                A Pandas dataFrame whose data will be plotted.

            methodName (optional)

                The name of the ParaMonte sample requesting the BasePlot.

            reportEnabled (optional)

                A boolean whose value indicates whether guidelines 
                should be printed in the standard output.

            resetPlot (optional)

                A function that resets the properties of the plot as desired 
                from outside. If provided, a pointer to this function will be
                saved for future internal usage.

        **Attributes**

            columns

                An attribute that determines the columns of dataFrame 
                to be visualized. It can have three forms:

                    1.  A list of column indices in dataFrame.
                    2.  A list of column names in dataFrame.columns.
                    3.  A ``range(start,stop,step)`` of column indices.

                Examples:

                    1.  ``xcolumns = [0,1,4,3]``
                    2.  ``xcolumns = ["SampleLogFunc","SampleVariable1"]``
                    3.  ``xcolumns = range(17,7,-2)``

                The default behavior includes all columns of the dataFrame.

            rows

                An attribute that determines the rows of dataFrame to be 
                visualized. It can be either:

                    1.  A ``range(start,stop,step)``, or, 
                    2.  A list of row indices in dataFrame.index.

                Examples:

                    1.  ``rows = range(17,7,-2)``
                    2.  ``rows = [i for i in range(7,17)]``

                The default behavior includes all rows of the dataFrame.

            heatmap

                A structure with one attribute:

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``heatmap()`` function of 
                        the seaborn library.

                        Example usage:

                            .. code-block:: python

                                heatmap.square = True

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            annotPrecision

                optional integer which determines the precision with which 
                 the output numbers corresponding to each pixel are written 
                 to the heatmap when annotation is requested via 
                 ``heatmap.kws.annot = True``. The default is 2, meaning 
                 that only two digits after the decimal are considered.

            xticklabels

                A structure with one attribute:

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``set_xticklabels()`` 
                        function of the matplotlib library.

                        Example usage:

                            .. code-block:: python

                                xticklabels.verticalalignment = "right"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            yticklabels

                A structure with one attribute:

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``set_yticklabels()`` 
                        function of the matplotlib library.

                        Example usage:

                            .. code-block:: python

                                yticklabels.verticalalignment = "right"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            set

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``set()``
                        function of the seaborn library should be made or not.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``set()`` function.

                        Example usage:

                            .. code-block:: python

                                set.kws.style = "darkgrid"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            axes (available only in 1D and 2D plots)

                A structure with one attribute:

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``gca()`` function of the 
                        matplotlib library.

                        Example usage:

                            .. code-block:: python

                                axes.kws.facecolor = "w"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            figure

                A structure with two attributes:

                    enabled

                        A boolean indicating whether a call to the ``figure()``
                        function of the matplotlib library should be made or not.
                        If a call is made, a new figure will be generated.
                        Otherwise, the current active figure will be used.

                    kws

                        A structure whose components are directly passed as 
                        keyword arguments to the ``figure()`` function.

                        Example usage:

                            .. code-block:: python

                                figure.kws.facecolor = "w"

                        **NOTE**

                        If a desired property is missing among the ``kws`` 
                        attributes, simply add the field and its value to 
                        the component.

            currentFig

                A structure whose attributes are the outputs of various plotting 
                tools used to make the current figure. These include the handle 
                to the current figure, the handle to the current axes in the plot, 
                the handle to the colorbar (if any exists), and other Python 
                plotting tools used to make to generate the figure.

        **Returns**

                An object of class ``HeatMapPlot``.

    ---------------------------------------------------------------------------
    """

    ################################################################################################################################
    #### __init__
    ################################################################################################################################

    def __init__( self
                , plotType      # : str
                , dataFrame     = None # : tp.Optional[ pd.DataFrame ] = None
                , methodName    = "ParaMonte" # : tp.Optional[ str ] = "ParaMonte"
                , reportEnabled = True # : tp.Optional[ bool ] = True
                , resetPlot     = None
                ):

        super().__init__( plotType = plotType
                        , dataFrame = dataFrame
                        , methodName = methodName
                        , reportEnabled = reportEnabled
                        , resetPlot = resetPlot
                        )

        self._reset()
        if resetPlot is None: self._resetPlot = self._reset
        self._progress.note()

    ################################################################################################################################
    #### _reset
    ################################################################################################################################

    def _reset(self):

        super()._reset()

        self.heatmap = Struct()
        self.heatmap.enabled = True
        self.heatmap.kws = Struct()

        self.xticklabels = Struct()
        self.xticklabels.kws = Struct()

        self.yticklabels = Struct()
        self.yticklabels.kws = Struct()

        self.columns = ""

        self._colorStart    = 20
        self._colorCount    = 200
        self._colorEnd      = 220

        self.annotPrecision = 2

        self._isdryrun = True
        self.make()
        self._isdryrun = False

    ################################################################################################################################
    #### __call__
    ################################################################################################################################

    def __call__( self
                , reself : tp.Optional[ bool ] = False
                , **kwargs
                ):
        """

        Call the ``make()`` method of the 
        current instance of the class.

            **Parameters**

                Any arguments that can be passed to the 
                ``make()`` method of the plot object.

            **Returns**

                Any return value from the ``make()`` method of the plot object.

        """

        return self.make(reself, **kwargs)

    ################################################################################################################################
    #### make
    ################################################################################################################################

    def make( self
            , reself : tp.Optional[ bool ] = False
            , **kwargs
            ):
        """

        Generate a heatmap plot from the selected 
        columns of the object's dataframe.

            **Parameters**

                reself

                    A logical variable. If ``True``, an instance of the 
                    object will be returned upon exit to the calling 
                    routine. The default value is ``False``.

            **Returns**

                the object self if ``reself = True`` otherwise, ``None``.
                However, this method causes side-effects by manipulating
                the existing attributes of the object.

        """

        for key in kwargs.keys():
            if hasattr(self,key):
                setattr(self, key, kwargs[key])
            elif key=="dataFrame":
                setattr( self, "_dfref", wref.ref(kwargs[key]) )
            else:
                raise Exception ( newline
                                + "Unrecognized input '"+key+"' class attribute detected." + newline
                                + self._getDocString()
                                )

        # set what to plot

        ############################################################################################################################
        #### xticklabels / yticklabels properties
        ############################################################################################################################

        if isinstance(self.xticklabels.kws,Struct):
            if "horizontalalignment" not in vars(self.xticklabels.kws).keys(): self.xticklabels.kws.horizontalalignment = "right"
            if "rotation" not in vars(self.xticklabels.kws).keys(): self.xticklabels.kws.rotation = 45
        else:
            raise Exception ( newline
                            + "The xticklabels.kws component of the current HeatMapPlot object must" + newline
                            + "be an object of class Struct(), essentially a structure with components" + newline
                            + "whose names are the input arguments to the set_xticklabels() method of the" + newline
                            + "Axes class of the matplotlib library." + newline
                            + self._getDocString()
                            )

        if isinstance(self.yticklabels.kws,Struct):
            if "horizontalalignment" not in vars(self.yticklabels.kws).keys(): self.yticklabels.kws.horizontalalignment = "right"
            if "rotation" not in vars(self.yticklabels.kws).keys(): self.yticklabels.kws.rotation = 45
        else:
            raise Exception ( newline
                            + "The yticklabels.kws component of the current HeatMapPlot object must" + newline
                            + "be an object of class Struct(), essentially a structure with components" + newline
                            + "whose names are the input arguments to the set_yticklabels() method of the" + newline
                            + "Axes class of the matplotlib library." + newline
                            + self._getDocString()
                            )

        ############################################################################################################################
        #### heatmap properties
        ############################################################################################################################

        if isinstance(self.heatmap.kws,Struct):
            if "square" not in vars(self.heatmap.kws).keys(): self.heatmap.kws.square = True
            if "cmap" not in vars(self.heatmap.kws).keys():
                try:
                    import seaborn as sns
                    self.heatmap.kws.cmap = sns.diverging_palette   ( h_neg = self._colorStart
                                                                    , h_pos = self._colorEnd
                                                                    , n = self._colorCount
                                                                    )
                except:
                    if self._isdryrun:
                        self.heatmap.kws.cmap = None
                    else:
                        raise Exception ( newline
                                        + "Failed to set the heatmap.kws.cmap component of the current HeatMapPlot object." + newline
                                        + "This component depends on the external seaborn Python library. Therefore, it is " + newline
                                        + "likely that the seaborn library or one of the required components of it, such as " + newline
                                        + "the matplotlib Python library is not properly installed on your system. Please " + newline
                                        + "fix this issue, otherwise, the visualization tools of the ParaMonte library " + newline
                                        + "will not work as expected. You can install the seaborn library by typing " + newline
                                        + "the following commands in your Anaconda3 or Bash command prompt: " + newline
                                        + newline
                                        + "    pip install --user --upgrade matplotlib"
                                        + "    pip install --user --upgrade seaborn"
                                        + newline
                                        + self._getDocString()
                                        )
        else:
            raise Exception ( newline
                            + "The heatmap.kws component of the current HeatMapPlot object must" + newline
                            + "be an object of class Struct(), essentially a structure with components" + newline
                            + "whose names are the input arguments to the heatmap() function of the" + newline
                            + "seaborn library." + newline
                            + self._getDocString()
                            )

        ############################################################################################################################
        #### figure properties
        ############################################################################################################################

        if self.figure.enabled:
            if isinstance(self.figure.kws, Struct):
                if "dpi" not in vars(self.figure.kws).keys(): self.figure.kws.dpi = 150
                if "facecolor" not in vars(self.figure.kws).keys(): self.figure.kws.facecolor = "w"
                if "edgecolor" not in vars(self.figure.kws).keys(): self.figure.kws.edgecolor = "w"
            else:
                raise Exception ( newline
                                + "The figure.kws component of the current DensityPlot object must" + newline
                                + "be an object of class Struct(), essentially a structure with components" + newline
                                + "whose names are the input arguments to the figure() function of the" + newline
                                + "matplotlib library." + newline
                                + self._getDocString()
                                )

        ############################################################################################################################
        ############################################################################################################################
        if self._isdryrun: return
        ############################################################################################################################
        ############################################################################################################################

        import seaborn as sns
        import matplotlib.pyplot as plt

        ############################################################################################################################
        #### generate figure and axes if needed
        ############################################################################################################################

        self._constructBasePlot()

        ############################################################################################################################
        #### check data type
        ############################################################################################################################

        self._checkDataType()

        ############################################################################################################################
        #### check rows presence. This must be checked here, because it depends on the integrity of the in input dataFrame.
        ############################################################################################################################

        if self.rows is None: self.rows = range(len(self._dfref().index))
        rownames = self._dfref().index[self.rows]

        ############################################################################################################################
        #### check columns presence. This must be checked here, because it depends on the integrity of the in input dataFrame.
        ############################################################################################################################

        colnames, colindex = pm.dfutils.getColNamesIndex(self._dfref().columns,self.columns)

        ############################################################################################################################
        #### set up tick labels
        ############################################################################################################################

        xtickExists = True
        if "xticklabels" in vars(self.heatmap.kws).keys():
            if not any(self.heatmap.kws.xticklabels): xtickExists = False
        else:
            self.heatmap.kws.xticklabels = colnames

        ytickExists = True
        if "yticklabels" in vars(self.heatmap.kws).keys():
            if not any(self.heatmap.kws.yticklabels): ytickEyists = False
        else:
            self.heatmap.kws.yticklabels = rownames

        ############################################################################################################################
        #### plot data
        ############################################################################################################################

        if self.annotPrecision is None:
            data = self._dfref().iloc[self.rows,colindex]
        else:
            data = self._dfref().iloc[self.rows,colindex].round( decimals = self.annotPrecision )

        self.currentFig.axes = sns.heatmap  ( data = data
                                            , **vars(self.heatmap.kws)
                                            )

        ############################################################################################################################
        #### configure the tick labels (orientation, ...)
        ############################################################################################################################

        self.currentFig.axes.set_xticklabels( self.currentFig.axes.get_xticklabels()
                                            , **vars(self.xticklabels.kws)
                                            );

        self.currentFig.axes.set_yticklabels( self.currentFig.axes.get_yticklabels()
                                            , **vars(self.yticklabels.kws)
                                            );

        plt.tight_layout()

        if self.figure.enabled:

            # default figure size

            figWidth = 6.4  # inches
            figHeight = 4.8 # inches
            figWidthScale = 1
            figHeightScale = 1
            threshDimension = 10

            # scale only if ticklabels are present

            if xtickExists:
                figWidthScale = max( 1 , self._dfref().shape[1]/threshDimension )
                figWidth *= figWidthScale

            if ytickExists:
                figHeightScale = max( 1 , self._dfref().shape[0]/threshDimension )
                figHeight *= figHeightScale

            self.currentFig.figure.set_size_inches(figWidth,figHeight)

        ############################################################################################################################

        if reself: return self

    ################################################################################################################################
    #### _getDocString
    ################################################################################################################################

    def _getDocString(self):
        docString   = newline \
                    + "Here is the help information on the HeatMapPlot class:" + newline \
                    + newline \
                    + self.__doc__ \
                    + newline \
                    + "Here is the help information on the parent BasePlot class:" + newline \
                    + newline \
                    + super().__doc__
        return docString

    ################################################################################################################################
    #### helpme
    ################################################################################################################################

    def helpme(self, topic=None):
        """

        Print the documentation for the input string topic. 
        If the topic does not exist, the documentation for
        the object will be printed.
     
            **Parameters**
         
                topic (optional)

                A string containing the name of the object 
                for which help is needed.

            **Returns**
         
                None
         
            **Example**

                .. code-block:: python
                    :linenos:

                    helpme()
                    helpme("make")
                    helpme("helpme")
                    helpme("getLogLinSpace")

        """

        try:
            exec("print(self."+topic+".__doc__)")
        except:
            print(self._getDocString())
        return None

    ################################################################################################################################



####################################################################################################################################
####################################################################################################################################
####
####   MIT License
####
####   ParaMonte: plain powerful parallel Monte Carlo library.
####
####   Copyright (C) 2012-present, The Computational Data Science Lab
####
####   This file is part of the ParaMonte library.
####
####   Permission is hereby granted, free of charge, to any person obtaining a 
####   copy of this software and associated documentation files (the "Software"), 
####   to deal in the Software without restriction, including without limitation 
####   the rights to use, copy, modify, merge, publish, distribute, sublicense, 
####   and/or sell copies of the Software, and to permit persons to whom the 
####   Software is furnished to do so, subject to the following conditions:
####
####   The above copyright notice and this permission notice shall be 
####   included in all copies or substantial portions of the Software.
####
####   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
####   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
####   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
####   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
####   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
####   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
####   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
####
####   ACKNOWLEDGMENT
####
####   ParaMonte is an honor-ware and its currency is acknowledgment and citations.
####   As per the ParaMonte library license agreement terms, if you use any parts of 
####   this library for any purposes, kindly acknowledge the use of ParaMonte in your 
####   work (education/research/industry/development/...) by citing the ParaMonte 
####   library as described on this page:
####
####       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
####
####################################################################################################################################
####################################################################################################################################

import sys
import numpy as np
from _pmutils import getList, getRandomFilePrefix

####################################################################################################################################
#### SpecBase specification type-checking class
####################################################################################################################################

inputFileDelim = ","

####################################################################################################################################

def sampleSize(sampleSize):
    if isinstance(sampleSize,(float,int)):
        return "sampleSize=" + str(sampleSize) + inputFileDelim
    else:
        raise TypeError("The input specification, sampleSize, must be of type float or int.")

####################################################################################################################################

def randomSeed(randomSeed):
    if isinstance(randomSeed,int):
        return "randomSeed=" + str(randomSeed) + inputFileDelim
    else:
        raise TypeError("The input specification, randomSeed, must be of type int.")

####################################################################################################################################

def description(description): return "description=" + verifyEncloseString(description,"description") + inputFileDelim

####################################################################################################################################

def outputFileName(outputFileName): return "outputFileName=" + verifyEncloseString(outputFileName,"outputFileName") + inputFileDelim

####################################################################################################################################

def outputDelimiter(outputDelimiter): return "outputDelimiter=" + verifyEncloseString(outputDelimiter,"outputDelimiter") + inputFileDelim

####################################################################################################################################

def chainFileFormat(chainFileFormat):  return "chainFileFormat=" + verifyEncloseString(chainFileFormat,"chainFileFormat") + inputFileDelim

####################################################################################################################################

def variableNameList(variableNameList):
    if isinstance(variableNameList,(list,tuple)):
        return "variableNameList=" + inputFileDelim.join("{0}".format(verifyEncloseString(_,"variableNameList["+str(i)+"]")) for i,_ in enumerate(getList(variableNameList))) + inputFileDelim
    else:
        raise TypeError("The input specification, variableNameList, must be either a list or tuple of ndim or less elements, each element of which must be of type str.")

####################################################################################################################################

def restartFileFormat(restartFileFormat): return "restartFileFormat=" + verifyEncloseString(restartFileFormat,"restartFileFormat") + inputFileDelim

####################################################################################################################################

def outputColumnWidth(outputColumnWidth):
    if isinstance(outputColumnWidth,int):
        return "outputColumnWidth=" + str(outputColumnWidth) + inputFileDelim
    else:
        raise TypeError("The input specification, outputColumnWidth, must be of type int.")

####################################################################################################################################

def overwriteRequested(overwriteRequested):
    if isinstance(overwriteRequested,bool):
        return "overwriteRequested=" + str(overwriteRequested) + inputFileDelim
    else:
        raise TypeError("The input specification, overwriteRequested, must be of type bool (True or False).")

####################################################################################################################################

def outputRealPrecision(outputRealPrecision):
    if isinstance(outputRealPrecision,int):
        return "outputRealPrecision=" + str(outputRealPrecision) + inputFileDelim
    else:
        raise TypeError("The input specification, outputRealPrecision, must be of type int.")

####################################################################################################################################

def silentModeRequested(silentModeRequested):
    if isinstance(silentModeRequested,bool):
        return "silentModeRequested=" + str(silentModeRequested) + inputFileDelim
    else:
        raise TypeError("The input specification, silentModeRequested, must be of type bool (True or False).")

####################################################################################################################################

def domainLowerLimitVec(domainLowerLimitVec):
    if isinstance(domainLowerLimitVec,(list,tuple,np.ndarray)):
        return "domainLowerLimitVec=" + str(np.array(getList(domainLowerLimitVec)).flatten()).strip('[]') + inputFileDelim
    else:
        raise TypeError("The input specification, domainLowerLimitVec, must be a list, tuple, or numpy vector of ndim or less elements of type float.")

####################################################################################################################################

def domainUpperLimitVec(domainUpperLimitVec):
    if isinstance(domainUpperLimitVec,(list,tuple,np.ndarray)):
        return "domainUpperLimitVec=" + str(np.array(getList(domainUpperLimitVec)).flatten()).strip('[]') + inputFileDelim
    else:
        raise TypeError("The input specification, domainUpperLimitVec, must be a list, tuple, or numpy vector of ndim or less elements of type float.")

####################################################################################################################################

def parallelizationModel(parallelizationModel):
    if isinstance(parallelizationModel,str):
        return "parallelizationModel=" + "'" + str(parallelizationModel) + "'" + inputFileDelim
    else:
        raise TypeError("The input specification, parallelizationModel, must be of type str.")

####################################################################################################################################

def progressReportPeriod(progressReportPeriod):
    if isinstance(progressReportPeriod,int):
        return "progressReportPeriod=" + str(progressReportPeriod) + inputFileDelim
    else:
        raise TypeError("The input specification, progressReportPeriod, must be of type int.")

####################################################################################################################################

def targetAcceptanceRate(targetAcceptanceRate):
    if isinstance(targetAcceptanceRate,(float,list,tuple,np.ndarray)):
        return "targetAcceptanceRate=" + str(np.array(getList(targetAcceptanceRate)).flatten()).strip('[]') + inputFileDelim
    else:
        raise TypeError("The input specification, targetAcceptanceRate, must be of type float.")

####################################################################################################################################

def mpiFinalizeRequested(mpiFinalizeRequested):
    if isinstance(mpiFinalizeRequested,bool):
        return "mpiFinalizeRequested=" + str(mpiFinalizeRequested) + inputFileDelim
    else:
        raise TypeError("The input specification, mpiFinalizeRequested, must be of type bool (True or False).")

####################################################################################################################################

def maxNumDomainCheckToWarn(maxNumDomainCheckToWarn):
    if isinstance(maxNumDomainCheckToWarn,int):
        return "maxNumDomainCheckToWarn=" + str(maxNumDomainCheckToWarn) + inputFileDelim
    else:
        raise TypeError("The input specification, maxNumDomainCheckToWarn, must be of type int.")

####################################################################################################################################

def maxNumDomainCheckToStop(maxNumDomainCheckToStop):
    if isinstance(maxNumDomainCheckToStop,int):
        return "maxNumDomainCheckToStop=" + str(maxNumDomainCheckToStop) + inputFileDelim
    else:
        raise TypeError("The input specification, maxNumDomainCheckToStop, must be of type int.")

####################################################################################################################################

def interfaceType():
    return "interfaceType=" + "'Python " + sys.version + "'" + inputFileDelim

####################################################################################################################################

def systemInfoFilePath(systemInfoFilePath):
    if systemInfoFilePath is None:
        return ""
    else:
        return "systemInfoFilePath='" + systemInfoFilePath + "'" + inputFileDelim

####################################################################################################################################
#### generate output filename
####################################################################################################################################

def genOutputFileName(methodName): return getRandomFilePrefix(prefix = methodName + "_run_")

####################################################################################################################################
#### enclose string with either single or double quotation . Return None if string contains both quotation symbols
####################################################################################################################################

def encloseString(string):
    hasSingleQuote = "'" in string
    hasDoubleQuote = '"' in string
    if hasSingleQuote:
        if hasDoubleQuote:
            return None
        else:
            return '"' + string + '"'
    else:
        return "'" + string + "'"

####################################################################################################################################
#### verify string value of the input variable does not contain both single and double quotations.
####################################################################################################################################

def verifyEncloseString(variableValue,variableName):
    if isinstance(variableValue,str):
        enclosedString = encloseString(variableValue)
        if enclosedString:
            return enclosedString
        else:
            raise TypeError ( "The input specification, " + variableName + ", cannot contain both single-quote and double-quote characters. "
                            + "Use only one type of quotation marks in your input string. " + variableName + " = "
                            + variableValue 
                            )
    else:
        raise TypeError("The input specification, " + variableName + ", must be of type str.")
